"""
VM network entity.
"""


class VMNetwork:
    def __init__(self, name, default=False):
        self.name = name
        self.default = default

        self.ips = []

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return isinstance(other, VMNetwork) and (self.name == other.name
                                                 and self.default == other.default)

    def __repr__(self):
        return f'''
                name: {self.name}
                default: {self.default}
                ips: {self.ips}
                '''

    def to_dict(self):
        return {
            'name': self.name,
            'default': self.default,
            'ips': self.ips,
        }



"""
Output messages and formatting.
"""

import sys
from datetime import datetime

from pytz import timezone

# Global variables
_SHOW_COLORS = True


def set_parametros_globais_escrita(no_color):
    global _SHOW_COLORS  # pylint: disable=global-statement
    _SHOW_COLORS = not no_color


def formatar_msg_aviso(msg):
    msg_inicial = 'WARNING'
    if _SHOW_COLORS:
        return f'\033[93m{msg_inicial}: {msg}\033[0m'
    return f'{msg_inicial}: {msg}'


def formatar_msg_erro(msg):
    if _SHOW_COLORS:
        return f'\033[91m{msg}\033[0m'
    return msg


def finalizar_com_erro(msg_erro):
    print(formatar_msg_erro(f'\nOperation error:\n{msg_erro}'))
    sys.exit(1)


def imprimir_ok(ocultar_progresso):
    if not ocultar_progresso:
        if _SHOW_COLORS:
            print('\033[92m[OK]\033[0m')
        else:
            print('[OK]')


def imprimir_erro(ocultar_progresso):
    if not ocultar_progresso:
        if _SHOW_COLORS:
            print('\033[91m[ERR]\033[0m')
        else:
            print('[ERR]')


def imprimir_acao_corrente(acao, ocultar_progresso, max_len=60):
    if not ocultar_progresso:
        print(f'{acao:<{max_len-3}}...', end='', flush=True)


def get_str_data_formatada(formato):
    fuso = timezone('America/Sao_Paulo')
    return datetime.now().astimezone(fuso).strftime(formato)



"""
Operation helper functions.
"""

import json
import os
import platform
import sys

from vmm_manager.infra.command import Command
from vmm_manager.util.msgs import (finalizar_com_erro, formatar_msg_aviso,
                                   formatar_msg_erro, get_str_data_formatada,
                                   imprimir_acao_corrente, imprimir_erro,
                                   imprimir_ok)


def __confirmar_acao_usuario(
    servidor_acesso=None,
    group=None,
    cloud=None,
    ocultar_progresso=False
):
    resposta = None
    while resposta not in ['s', 'n', 'y']:
        resposta = input('Continue? [y/n] ').lower()

        if resposta == 'n':
            if servidor_acesso and group and cloud:
                remove_operation_lock(servidor_acesso, group,
                                      cloud, ocultar_progresso)

            if not ocultar_progresso:
                print(formatar_msg_aviso('Canceled by user.'))

            sys.exit(0)


def confirmar_acao_usuario_sem_lock():
    __confirmar_acao_usuario()


def confirmar_acao_usuario_com_lock(servidor_acesso, group, cloud, ocultar_progresso):
    __confirmar_acao_usuario(
        servidor_acesso, group, cloud, ocultar_progresso)


def validar_retorno_operacao_sem_lock(status, msg, ocultar_progresso):
    if status:
        imprimir_ok(ocultar_progresso)
    else:
        imprimir_erro(ocultar_progresso)
        finalizar_com_erro(msg)


def validar_retorno_operacao_com_lock(
    status,
    msg,
    servidor_acesso,
    group,
    cloud,
    ocultar_progresso
):
    if status:
        imprimir_ok(ocultar_progresso)
    else:
        imprimir_erro(ocultar_progresso)
        remove_operation_lock(servidor_acesso, group, cloud, ocultar_progresso)
        finalizar_com_erro(msg)


def add_operation_lock(servidor_acesso, group, cloud, ocultar_progresso):
    imprimir_acao_corrente('Adding lock', ocultar_progresso)

    cmd = Command('add_operation_lock',
                  lockfile=servidor_acesso.get_caminho_lockfile(
                      group,
                      cloud),
                  pid=os.getpid(),
                  host=platform.node(),
                  group=group,
                  cloud=cloud,
                  data=get_str_data_formatada('%m-%d-%Y at %I:%M:%S.%f %p'))

    status, retorno_cmd = cmd.executar(servidor_acesso)
    if status:
        dados_lock = json.loads(retorno_cmd)
        if not dados_lock.get('Success'):
            status = False
            retorno_cmd = f"The process {dados_lock.get('PID')}, " \
                f"started in {dados_lock.get('DataLock')} " \
                f"on the server {dados_lock.get('HostLock')}, " \
                f"is already working on the {dados_lock.get('Group')} group " \
                f"in the {dados_lock.get('Cloud')} cloud."

    validar_retorno_operacao_sem_lock(status, retorno_cmd, ocultar_progresso)


def remove_operation_lock(servidor_acesso, group, cloud, ocultar_progresso):
    imprimir_acao_corrente('Removing lock', ocultar_progresso)

    cmd = Command('remove_operation_lock',
                  lockfile=servidor_acesso.get_caminho_lockfile(
                      group,
                      cloud))

    status, _ = cmd.executar(servidor_acesso)
    if status:
        imprimir_ok(ocultar_progresso)
    else:
        imprimir_erro(ocultar_progresso)
        print(formatar_msg_erro(
            'The lock file could not be removed. Please remove it manually.'))



"""
VMM Manager enums.
"""

from enum import Enum


class VMStatusEnum(Enum):
    RUNNING = 0
    POWER_OFF = 1
    UPDATED_WITH_ERROR = 107
    INCOMPLETE_VM_CONFIGURATION = 223


class SCJobStatusEnum(Enum):
    RUNNING = 1
    FAILED = 3
    CANCELED = 4
    SUCCESS = 5
    SUCCESS_WITH_WARNING = 6


class SCDiskBusType(Enum):
    SCSI = 'SCSI'
    IDE = 'IDE'


class SCDiskSizeType(Enum):
    FIXED = 'FixedSize'
    DYNAMIC = 'DynamicallyExpanding'



"""
Inventory entity.
"""
import json

from vmm_manager.entity.plan import Plan
from vmm_manager.infra.command import Command
from vmm_manager.scvmm.scregion import SCRegion
from vmm_manager.util.config import FIELD_GROUP


def json_handle_inventory(obj):
    if isinstance(obj, Inventory):
        return obj.to_dict()

    raise ValueError('Object must be an Inventory')


class Inventory:

    @staticmethod
    def get_json(inventario_local, inventario_remoto, all_data=True):
        for vm_name in inventario_remoto.vms:
            if vm_name not in inventario_local.vms:
                # máquina órfã: não exibir
                continue

            inventario_remoto.vms[vm_name].dados_ansible = \
                inventario_local.vms[vm_name].dados_ansible

            # definindo tipo de impressão
            inventario_remoto.vms[vm_name].to_json_dados_completos = all_data

        return True, json.dumps(inventario_remoto,
                                default=json_handle_inventory,
                                sort_keys=True, indent=4)

    def __init__(self, group, cloud):
        self.group = group
        self.cloud = cloud
        self.vms = {}
        self.interval_between_resources = 0

        self.__regioes_por_letra_id = None
        self.__regioes_disponiveis = None

    def get_nome_no_regiao(self, region):
        self.__retirar_regiao_pool(region)

        if region in self.__regioes_por_letra_id:
            return self.__regioes_por_letra_id[region].nome_no

        raise ValueError(f"Region '{region}' does not have a host defined.")

    def get_id_no_regiao(self, region):
        self.__retirar_regiao_pool(region)

        if region in self.__regioes_por_letra_id:
            return self.__regioes_por_letra_id[region].id_no

        raise ValueError(f"Region '{region}' does not have a host defined.")

    def get_mapeamento_regioes_to_test(self):
        # flush regions
        it_regioes = 0
        while self.__regioes_disponiveis:
            self.__retirar_regiao_pool(chr(ord('A') + it_regioes))
            it_regioes += 1

        if self.__regioes_por_letra_id is None:
            raise ValueError('Regions map not set.')

        return self.__regioes_por_letra_id

    def set_regioes_disponiveis(self, regioes_disponiveis):
        self.__regioes_por_letra_id = {}
        self.__regioes_disponiveis = regioes_disponiveis

    def __retirar_regiao_pool(self, region):
        if region in self.__regioes_por_letra_id:
            return

        if not self.__regioes_disponiveis:
            return

        self.__regioes_por_letra_id[region] = self.__regioes_disponiveis.pop()

    def calcular_plano_execucao(self, inventario_remoto):
        if (self.group != inventario_remoto.group
                or self.cloud != inventario_remoto.cloud):
            return False, 'Inventories must be from the same group and cloud.'

        plano_execucao = Plan(self.group, self.cloud)

        # update interval between resources
        plano_execucao.interval_between_resources = self.interval_between_resources

        self.__add_acoes_criar_vms(inventario_remoto, plano_execucao)
        self.__add_acoes_execucao_excluir_vms(
            inventario_remoto, plano_execucao)

        self.__add_acoes_diferenca_vm(
            inventario_remoto, plano_execucao)
        self.__add_acoes_diferenca_discos_adicionais(
            inventario_remoto, plano_execucao)
        self.__add_acoes_diferenca_regiao(
            inventario_remoto, plano_execucao)
        self.__add_acoes_virtualizacao_aninhada(
            inventario_remoto, plano_execucao)
        self.__add_acoes_memoria_dinamica(
            inventario_remoto, plano_execucao)

        return True, plano_execucao

    def gerar_plano_exclusao(self):
        plano_execucao = Plan(self.group, self.cloud)

        for maquina_virtual in self.vms.values():
            plano_execucao.actions.append(
                maquina_virtual.get_acao_excluir_vm())

        return plano_execucao

    def is_vazio(self):
        return not self.vms

    def lista_nome_vms_str(self):
        return ','.join([f'"{vm_name}"' for vm_name in self.vms])

    def __validar_regras_locais(self):
        regioes = set()

        for maquina_virtual in self.vms.values():
            if maquina_virtual.region != SCRegion.REGION_DEFAULT:
                regioes.add(maquina_virtual.region)

            if maquina_virtual.image is None:
                raise ValueError(
                    f'VM {maquina_virtual.name} does not have an image defined.')

            if maquina_virtual.cpu is None:
                raise ValueError(
                    f'VM {maquina_virtual.name} does not have a CPU defined.')

            if maquina_virtual.memory is None:
                raise ValueError(
                    f'VM {maquina_virtual.name} does not have a memory defined.')

            if maquina_virtual.get_qtde_rede_principal() != 1:
                raise ValueError(
                    f'VM {maquina_virtual.name} should have only one primary network defined.')

    def validar(self, servidor_acesso):
        self.__validar_regras_locais()

        imagens = set()
        networks = set()
        regioes = set()

        for maquina_virtual in self.vms.values():
            imagens.add(maquina_virtual.image)

            if maquina_virtual.region != SCRegion.REGION_DEFAULT:
                regioes.add(maquina_virtual.region)

            networks.update(
                [network.name for network in maquina_virtual.networks])

        cmd = Command(
            'validate_inventory', imagens=imagens,
            cloud=self.cloud,
            networks=networks,
            vmm_server=servidor_acesso.vmm_server,
            qtde_minima_regioes=len(regioes),
            group=self.group,
            lista_nome_vms_str=self.lista_nome_vms_str(),
            field_group=FIELD_GROUP[0]
        )
        _, msg = cmd.executar(servidor_acesso)
        if msg:
            raise ValueError(msg)

    def set_discos_vms(self, additional_disks):
        for vm_name in additional_disks:
            self.vms[vm_name].add_discos_adicionais(additional_disks[vm_name])

    def __add_acoes_criar_vms(self, inventario_remoto, plano_execucao):
        vms_inserir = [
            nome_vm_local for nome_vm_local in self.vms
            if nome_vm_local not in inventario_remoto.vms
        ]
        for vm_name in vms_inserir:
            plano_execucao.actions.append(
                self.vms[vm_name].get_acao_criar_vm())

    def __add_acoes_execucao_excluir_vms(self, inventario_remoto, plano_execucao):
        vms_excluir = [
            nome_vm_remoto for nome_vm_remoto in inventario_remoto.vms
            if nome_vm_remoto not in self.vms
        ]
        for vm_name in vms_excluir:
            plano_execucao.actions.append(
                inventario_remoto.vms[vm_name].get_acao_excluir_vm())

    def __add_acoes_diferenca_discos_adicionais(self, inventario_remoto, plano_execucao):
        for vm_name, data_vm in self.vms.items():
            data_vm.add_acoes_diferenca_discos_adicionais(
                inventario_remoto.vms.get(vm_name, None),
                plano_execucao)

    def __add_acoes_diferenca_regiao(self, inventario_remoto, plano_execucao):
        for vm_name, data_vm in self.vms.items():
            data_vm.add_acoes_diferenca_regiao(
                inventario_remoto.vms.get(vm_name, None),
                plano_execucao, inventario_remoto)

    def __add_acoes_virtualizacao_aninhada(self, inventario_remoto, plano_execucao):
        for vm_name, data_vm in self.vms.items():
            data_vm.add_acoes_virtualizacao_aninhada(
                inventario_remoto.vms.get(vm_name, None),
                plano_execucao)

    def __add_acoes_memoria_dinamica(self, inventario_remoto, plano_execucao):
        for vm_name, data_vm in self.vms.items():
            data_vm.add_acoes_memoria_dinamica(
                inventario_remoto.vms.get(vm_name, None),
                plano_execucao)

    def __add_acoes_diferenca_vm(self, inventario_remoto, plano_execucao):
        for vm_name, data_vm in self.vms.items():
            if vm_name in inventario_remoto.vms:
                data_vm.add_acoes_diferenca_vm(
                    inventario_remoto.vms[vm_name], plano_execucao)

    def __eq__(self, other):
        return isinstance(other, Inventory) and (self.group == other.group
                                                 and self.cloud == other.cloud
                                                 and self.vms == other.vms)

    def __str__(self):
        return f'''
            group: {self.group}
            cloud: {self.cloud}
            vms: {self.vms}
            '''

    def to_dict(self):
        return {
            'group': self.group,
            'cloud': self.cloud,
            'vms': [vm_data.to_dict() for vm_data in self.vms.values()]
        }



"""
Parser for local inventory file.
"""

import os

import yamale

from vmm_manager.entity.inventory import Inventory
from vmm_manager.entity.vm import VM
from vmm_manager.entity.vm_disk import VMDisk
from vmm_manager.entity.vm_network import VMNetwork
from vmm_manager.scvmm.enums import SCDiskBusType, SCDiskSizeType
from vmm_manager.scvmm.scregion import SCRegion


# pylint: disable=too-few-public-methods
class ParserLocal:
    __ARQUIVO_SCHEMA = '../includes/schema.yaml'
    __YAML_PARSER = 'ruamel'

    @staticmethod
    def __get_schema_yaml():
        return yamale.make_schema(
            os.path.join(
                os.path.dirname(__file__), ParserLocal.__ARQUIVO_SCHEMA),
            parser=ParserLocal.__YAML_PARSER)

    @staticmethod
    def __get_discos_adicionais(vm_name, dict_discos_adicionais):
        additional_disks = []
        for item in dict_discos_adicionais or {}:
            file = item.get('file')

            if file in [additional_disk.file for additional_disk in additional_disks]:
                raise ValueError(
                    f"Disk '{file}' already exists in VM '{vm_name}'.")

            additional_disk = VMDisk(
                SCDiskBusType(item.get('bus_type')),
                item.get('file'),
                item.get('size_mb'),
                SCDiskSizeType(item.get('size_type')),
                item.get('path'))

            additional_disks.append(additional_disk)

        return additional_disks

    def __init__(self, inventory_file):
        self.__arquivo_inventario = inventory_file
        self.__inventario = None

    def __validar_arquivo_yaml(self):
        if not os.path.isfile(self.__arquivo_inventario):
            raise ValueError('Inventory file not found.')
        if os.stat(self.__arquivo_inventario).st_size == 0:
            raise ValueError('Empty inventory file.')

    def __montar_inventario(
        self,
        dados_inventario,
        filtro_nome_vm=None,
        filtro_dados_completos=True
    ):
        nomes_vm = []
        self.__inventario = Inventory(
            dados_inventario['group'],
            dados_inventario['cloud']
        )

        # set interval between resources
        self.__inventario.interval_between_resources = dados_inventario.get(
            'interval_between_resources', 0
        )

        for maquina_virtual in dados_inventario['vms']:
            vm_name = maquina_virtual.get('name').upper()
            if vm_name in nomes_vm:
                raise ValueError(
                    f'VM {vm_name} already exists in inventory.')
            nomes_vm.append(vm_name)

            # filtrando vms: melhoria no desempenho
            if filtro_nome_vm and vm_name != filtro_nome_vm:
                continue

            vm_redes = []
            nomes_redes = []
            for rede_vm in maquina_virtual.get(
                'networks',
                dados_inventario.get('networks_default', [])
            ):
                nome_rede = rede_vm.get('name')

                if nome_rede in nomes_redes:
                    raise ValueError(
                        f"Network '{nome_rede}' already exists in VM '{vm_name}'.")

                nomes_redes.append(nome_rede)
                vm_redes.append(
                    VMNetwork(nome_rede, rede_vm.get('default', False)))

            self.__inventario.vms[vm_name] = VM(
                vm_name,
                maquina_virtual.get('description'),
                maquina_virtual.get(
                    'image', dados_inventario.get('image_default', None)),
                maquina_virtual.get('region', SCRegion.REGION_DEFAULT),
                maquina_virtual.get(
                    'cpu', dados_inventario.get('cpu_default', None)),
                maquina_virtual.get(
                    'memory', dados_inventario.get('memory_default', None)),
                vm_redes,
                nested_virtualization=maquina_virtual.get(
                    'nested_virtualization',
                    dados_inventario.get('nested_virtualization_default', False)),
                dynamic_memory=maquina_virtual.get(
                    'dynamic_memory',
                    dados_inventario.get('dynamic_memory_default', True)),
            )
            self.__inventario.vms[vm_name].extrair_dados_ansible_dict(
                maquina_virtual.get('ansible'))

            # Obtendo dados adicionais
            if filtro_dados_completos:
                # discos
                self.__inventario.vms[vm_name].add_discos_adicionais(
                    ParserLocal.__get_discos_adicionais(vm_name,
                                                        maquina_virtual.get('additional_disks')))

    def __carregar_yaml(self):
        return yamale.make_data(self.__arquivo_inventario,
                                parser=ParserLocal.__YAML_PARSER)

    def get_inventario(self, servidor_acesso, filtro_nome_vm=None, filtro_dados_completos=True):
        if not self.__inventario:
            try:
                self.__validar_arquivo_yaml()
                dados_yaml = self.__carregar_yaml()
                parser = ParserLocal.__get_schema_yaml()

                yamale.validate(parser, dados_yaml, strict=True)

                self.__montar_inventario(
                    dados_yaml[0][0], filtro_nome_vm, filtro_dados_completos)
                self.__inventario.validar(servidor_acesso)
            except (SyntaxError, ValueError) as ex:
                return False, str(ex)

        return True, self.__inventario



"""
Manages resources in the System Center Virtual Machine Manager (SCVMM),
in a declarative way, based on a YAML configuration file.
"""

import argparse
import os
import re
import string

import configargparse
from ruamel.yaml import YAML, scanner

from vmm_manager.entity.inventory import Inventory
from vmm_manager.entity.plan import Plan
from vmm_manager.infra.access_server import AccessServer
from vmm_manager.infra.command import Command
from vmm_manager.parser.parser_local import ParserLocal
from vmm_manager.parser.parser_remote import ParserRemote
from vmm_manager.util.config import (FIELD_GROUP, FIELD_ID, FIELD_IMAGE,
                                     FIELD_NETWORK_DEFAULT, FIELD_REGION)
from vmm_manager.util.msgs import (finalizar_com_erro, formatar_msg_aviso,
                                   imprimir_acao_corrente,
                                   set_parametros_globais_escrita)
from vmm_manager.util.operation import (add_operation_lock,
                                        confirmar_acao_usuario_com_lock,
                                        remove_operation_lock,
                                        validar_retorno_operacao_com_lock,
                                        validar_retorno_operacao_sem_lock)


def parametro_arquivo_yaml(nome_arquivo):
    try:
        with open(nome_arquivo, 'r', encoding='utf8') as stream:
            yaml = YAML()
            yaml.load(stream)

            return nome_arquivo
    except (FileNotFoundError, IsADirectoryError) as exc:
        raise argparse.ArgumentTypeError(exc)
    except scanner.ScannerError as exc:
        raise argparse.ArgumentTypeError(
            f'Invalid YAML: {exc}')


def parametro_alfanumerico_limitado(value):
    regex_alfa_num = re.compile(r'^[a-z]{1}[a-z0-9_]{2,39}$', re.IGNORECASE)

    if value and not regex_alfa_num.match(value):
        raise argparse.ArgumentTypeError(
            f"Invalid parameter: '{value}'")

    return value


def get_parser():
    parser = configargparse.ArgumentParser(
        description='''
        Python script that manages resources in the System Center Virtual Machine Manager (SCVMM), \
        in a declarative way, based on a YAML configuration file.
        ''', default_config_files=['vmm_manager.ini'])

    parser.add('-a', '--access-point',
               help='''
                Address (FQDN or IP) of the Windows server, \
                with the VirtualMachineManager PowerShell module installed, \
                which will run the PowerShell scripts. \
                E.g.: windows_host.domain.com
               ''',
               env_var='VMM_ACCESS_POINT', required=True, type=str)
    parser.add('-u', '--username',
               help='Username with access to the access point and the SCVMM',
               env_var='VMM_USERNAME', required=True, type=str)
    parser.add('-p', '--password',
               help='User password',
               env_var='VMM_PASSWORD', required=True, type=str)
    parser.add('-s', '--server',
               help='SCVMM server. E.g.: scvmm_server.domain.com',
               env_var='VMM_SERVER', required=True, type=str)
    parser.add('--hide-progress',
               help='Hide progress messages',
               env_var='VMM_HIDE_PROGRESS', required=False, action='store_true')
    parser.add('--no-color',
               help='Do not use colors in the output',
               env_var='VMM_NO_COLOR', required=False, action='store_true')

    subprasers = parser.add_subparsers(dest='command')
    plan = subprasers.add_parser(
        'plan', help='Create an execution plan \
            based on the difference between the local and remote inventory')
    plan.add_argument('--inventory',
                      help='YAML file with resource specifications',
                      dest='inventory_file', env_var='VMM_INVENTORY',
                      required=True, type=parametro_arquivo_yaml)

    apply = subprasers.add_parser(
        'apply',
        help='Apply the execution plan to the SCVMM')
    apply.add_argument('--execution-plan',
                       help='YAML file with the execution plan',
                       dest='execution_plan_file', env_var='VMM_EXECUTION_PLAN',
                       required=False, type=parametro_arquivo_yaml)
    apply.add_argument('--inventory',
                       help='YAML file with resource specifications',
                       dest='inventory_file', env_var='VMM_INVENTORY',
                       required=False, type=parametro_arquivo_yaml)
    apply.add_argument('--skip-confirmation',
                       help='Skip confirmation',
                       env_var='VMM_SKIP_CONFIRMATION', required=False, action='store_true')
    apply.add_argument('--interval-between-resources',
                       help='Interval (in seconds) to wait between operations on different resources. '
                            'It has precedence over the value defined in the inventory file. '
                            'It is useful when you are working with clusters, and you want to '
                            'wait for the node stabilization before starting the operations on the next node.',
                       env_var='VMM_INTERVAL_BETWEEN_RESOURCES', required=False, type=int, default=-1)

    destroy = subprasers.add_parser(
        'destroy', help='Remove all machines from the specified group in the specified cloud')
    destroy.add_argument(
        '--group', help='Group name', required=True)
    destroy.add_argument(
        '--cloud', help='Cloud name', required=True)
    destroy.add_argument('--skip-confirmation',
                         help='Skip confirmation',
                         env_var='VMM_SKIP_CONFIRMATION', required=False, action='store_true')

    subprasers.add_parser(
        'opts', help='List available options in the SCVMM for each field')

    show = subprasers.add_parser(
        'show', help='Show a JSON data with information about the resources')
    show.add_argument('--inventory',
                      help='YAML file with resource specifications',
                      dest='inventory_file', env_var='VMM_INVENTORY',
                      required=True, type=parametro_arquivo_yaml)
    show.add_argument('--vm-name',
                      help='Virtual machine name', default='',
                      required=False, type=parametro_alfanumerico_limitado)
    show.add_argument('--all-data',
                      help='Show all the resource data',
                      env_var='VMM_ALL_DATA', required=False, action='store_true')

    return parser


def obter_inventario_remoto(
    servidor_acesso,
    group,
    cloud,
    ocultar_progresso,
    filtro_nome_vm=None,
    filtro_dados_completos=True
):
    imprimir_acao_corrente('Loading remote inventory', ocultar_progresso)

    parser_remoto = ParserRemote(group, cloud)
    status, inventario_remoto = parser_remoto.get_inventario(
        servidor_acesso, filtro_nome_vm, filtro_dados_completos)
    validar_retorno_operacao_com_lock(status, inventario_remoto,
                                      servidor_acesso, group,
                                      cloud, ocultar_progresso)

    return inventario_remoto


def obter_inventario_local(
    servidor_acesso,
    inventory_file,
    ocultar_progresso,
    filtro_nome_vm=None,
    filtro_dados_completos=True
):
    imprimir_acao_corrente('Loading local inventory', ocultar_progresso)

    parser_local = ParserLocal(inventory_file)
    status, inventario_local = parser_local.get_inventario(
        servidor_acesso, filtro_nome_vm, filtro_dados_completos)
    validar_retorno_operacao_sem_lock(
        status, inventario_local, ocultar_progresso)

    return inventario_local


def obter_plano_execucao(servidor_acesso, inventario_local: Inventory, ocultar_progresso):
    inventario_remoto = obter_inventario_remoto(
        servidor_acesso, inventario_local.group,
        inventario_local.cloud, ocultar_progresso)

    imprimir_acao_corrente('Calculation execution plan', ocultar_progresso)
    status, plano_execucao = inventario_local.calcular_plano_execucao(
        inventario_remoto)
    validar_retorno_operacao_com_lock(status, plano_execucao, servidor_acesso,
                                      inventario_local.group,
                                      inventario_local.cloud, ocultar_progresso)
    return plano_execucao


def validar_conexao(servidor_acesso, ocultar_progresso):
    imprimir_acao_corrente('Checking VMM connection', ocultar_progresso)
    cmd = Command('connection_test', vmm_server=servidor_acesso.vmm_server)
    status, msg = cmd.executar(servidor_acesso)

    if status and 'True' not in msg:
        status = False
    validar_retorno_operacao_sem_lock(status, msg, ocultar_progresso)


def vmm_setup(servidor_acesso, ocultar_progresso):
    imprimir_acao_corrente('Setting up the VMM', ocultar_progresso)
    cmd = Command('vmm_setup', vmm_server=servidor_acesso.vmm_server,
                  campos_customizados=[FIELD_GROUP, FIELD_ID,
                                       FIELD_IMAGE, FIELD_REGION,
                                       FIELD_NETWORK_DEFAULT])
    status, msg = cmd.executar(servidor_acesso)

    validar_retorno_operacao_sem_lock(status, msg, ocultar_progresso)


def list_options(servidor_acesso, ocultar_progresso):
    imprimir_acao_corrente('Listing options', ocultar_progresso)
    cmd = Command('list_options', vmm_server=servidor_acesso.vmm_server)
    status, opcoes = cmd.executar(servidor_acesso)

    validar_retorno_operacao_sem_lock(status, opcoes, ocultar_progresso)
    regioes = re.search(r'regions: ([0-9]{1,})', opcoes).group(1)
    str_regioes = '\n'.join(string.ascii_uppercase[0:int(regioes)])
    opcoes = opcoes.replace(f'regions: {regioes}', str_regioes)
    print('\n' + opcoes)


def imprimir_json_inventario(
    servidor_acesso,
    inventory_file,
    vm_name,
    all_data,
    ocultar_progresso
):
    vmm_setup(servidor_acesso, ocultar_progresso)
    inventario_local = obter_inventario_local(
        servidor_acesso, inventory_file, ocultar_progresso,
        filtro_nome_vm=vm_name, filtro_dados_completos=all_data)

    add_operation_lock(servidor_acesso, inventario_local.group,
                       inventario_local.cloud, ocultar_progresso)

    inventario_remoto = obter_inventario_remoto(
        servidor_acesso, inventario_local.group,
        inventario_local.cloud, ocultar_progresso,
        filtro_nome_vm=vm_name, filtro_dados_completos=all_data)

    remove_operation_lock(servidor_acesso, inventario_local.group,
                          inventario_local.cloud, ocultar_progresso)

    imprimir_acao_corrente('Generating JSON', ocultar_progresso)
    status, json_inventario = Inventory.get_json(
        inventario_local, inventario_remoto, all_data)
    validar_retorno_operacao_sem_lock(
        status, json_inventario, ocultar_progresso)
    print(json_inventario)


def planejar_sincronizacao(servidor_acesso, inventory_file, ocultar_progresso):
    vmm_setup(servidor_acesso, ocultar_progresso)
    inventario_local = obter_inventario_local(
        servidor_acesso, inventory_file, ocultar_progresso)

    add_operation_lock(servidor_acesso,
                       inventario_local.group,
                       inventario_local.cloud, ocultar_progresso)

    conteudo_arquivo = None
    plano_execucao = obter_plano_execucao(
        servidor_acesso, inventario_local, ocultar_progresso)

    if not plano_execucao.is_vazio():
        imprimir_acao_corrente(
            f'Saving file {Plan.ARQUIVO_PLANO_EXECUCAO}',
            ocultar_progresso)
        status, conteudo_arquivo = plano_execucao.gerar_arquivo()
        validar_retorno_operacao_sem_lock(
            status, conteudo_arquivo, ocultar_progresso)

    remove_operation_lock(servidor_acesso, inventario_local.group,
                          inventario_local.cloud, ocultar_progresso)

    if conteudo_arquivo:
        print(f'\nChanges to apply:\n{conteudo_arquivo}')
    else:
        Plan.excluir_arquivo()
        print(
            '\nNo differences between the local and remote inventory: nothing to do.')


def executar_sincronizacao(
    servidor_acesso,
    execution_plan_file,
    skip_confirmation,
    inventory_file,
    ocultar_progresso,
    interval_between_resources
):
    vmm_setup(servidor_acesso, ocultar_progresso)

    # Obtendo plano de execução
    #
    # Caso de ter informado o plano de execução
    adquiriu_lock = False
    if execution_plan_file or os.path.isfile(Plan.ARQUIVO_PLANO_EXECUCAO):
        imprimir_acao_corrente(
            'Loading execution plan', ocultar_progresso)
        status, plano_execucao = Plan.carregar_plano_execucao(
            execution_plan_file or Plan.ARQUIVO_PLANO_EXECUCAO)
        validar_retorno_operacao_sem_lock(
            status, plano_execucao, ocultar_progresso)
    # Caso de ter informado o arquivo de inventário
    elif inventory_file:
        print(formatar_msg_aviso(
            'The execution plan will be generated based on the local inventory.'),
            flush=True)
        inventario_local = obter_inventario_local(
            servidor_acesso, inventory_file, ocultar_progresso)

        add_operation_lock(servidor_acesso, inventario_local.group,
                           inventario_local.cloud, ocultar_progresso)
        adquiriu_lock = True

        plano_execucao = obter_plano_execucao(
            servidor_acesso,
            inventario_local,
            ocultar_progresso
        )
    # Se não informou nem o plano nem o inventário, então informar erro
    else:
        finalizar_com_erro(
            'You must pass the execution plan or the inventory file.')

    if not adquiriu_lock:
        add_operation_lock(servidor_acesso, plano_execucao.group,
                           plano_execucao.cloud, ocultar_progresso)

    if plano_execucao.is_vazio():
        remove_operation_lock(servidor_acesso, plano_execucao.group,
                              plano_execucao.cloud, ocultar_progresso)
        print(
            '\nThe execution plan is empty: nothing to do.')
    else:
        # Exibindo confirmação
        if not skip_confirmation:
            print('\nThe following actions will be performed:')
            plano_execucao.imprimir_acoes()
            confirmar_acao_usuario_com_lock(
                servidor_acesso, plano_execucao.group,
                plano_execucao.cloud, ocultar_progresso)

        plano_execucao.executar(servidor_acesso, ocultar_progresso, interval_between_resources)

        remove_operation_lock(servidor_acesso, plano_execucao.group,
                              plano_execucao.cloud, ocultar_progresso)
        plano_execucao.imprimir_resultado_execucao()


def remover_agrupamento_da_nuvem(
    servidor_acesso,
    group,
    cloud,
    skip_confirmation,
    ocultar_progresso
):
    vmm_setup(servidor_acesso, ocultar_progresso)

    add_operation_lock(servidor_acesso, group, cloud, ocultar_progresso)

    inventario_remoto = obter_inventario_remoto(
        servidor_acesso, group, cloud, ocultar_progresso)

    if not inventario_remoto.is_vazio():
        # Exibindo confirmação
        if not skip_confirmation:
            print('\nThe following resources will be removed:')
            for maquina_virtual in inventario_remoto.vms.values():
                print(
                    f'VMM ID: {maquina_virtual.vmm_id}\t'
                    f'Name: {maquina_virtual.name}\t'
                    f'Status: {maquina_virtual.status}')
            confirmar_acao_usuario_com_lock(
                servidor_acesso, group, cloud, ocultar_progresso)

        plano_execucao = inventario_remoto.gerar_plano_exclusao()
        plano_execucao.executar(servidor_acesso, ocultar_progresso)

        remove_operation_lock(servidor_acesso, group, cloud, ocultar_progresso)
        plano_execucao.imprimir_resultado_execucao()
    else:
        remove_operation_lock(servidor_acesso, group, cloud, ocultar_progresso)
        print(
            f'\nThere are no resources in the cloud {cloud} for the group {group}.')


def main():
    parser = get_parser()
    args = parser.parse_args()

    set_parametros_globais_escrita(args.no_color)

    if not args.command:
        finalizar_com_erro('No command. Use -h for help.')

    servidor_acesso = AccessServer(
        args.access_point, args.username, args.password, args.server)
    validar_conexao(servidor_acesso, args.hide_progress)

    if args.command == 'plan':
        planejar_sincronizacao(
            servidor_acesso, args.inventory_file, args.hide_progress)
    elif args.command == 'apply':
        executar_sincronizacao(
            servidor_acesso, args.execution_plan_file,
            args.skip_confirmation, args.inventory_file,
            args.hide_progress, args.interval_between_resources)
    elif args.command == 'destroy':
        remover_agrupamento_da_nuvem(
            servidor_acesso, args.group, args.cloud,
            args.skip_confirmation, args.hide_progress)
    elif args.command == 'opts':
        list_options(servidor_acesso, args.hide_progress)
    elif args.command == 'show':
        imprimir_json_inventario(
            servidor_acesso, args.inventory_file,
            args.vm_name.upper(), args.all_data,
            args.hide_progress)
